;*********************************************************************
; 3031 CONTROLLER ~ MIDI to CV Converter                             *
;*********************************************************************
; (c) 2005 SM-Sound / Seb Francis                                    *
;*********************************************************************
;
; Main code
;
;*********************************************************************
;                                                               
;    Connections:
;
;    RA0   <- Noise
;    RA1   -> MUX0
;    RA2   -> MUX1
;    RA3   -> MUX2
;    RA4   Open drain output!
;    RA5   -> MUX3
;
;    RB0   -> NC8
;    RB1   -> NC9
;    RB2   -> NC10
;    RB3
;    RB4   <- MIDI Channel Select bit0
;    RB5   <- MIDI Channel Select bit1
;    RB6   <- MIDI Channel Select bit2
;    RB7   <- MIDI Channel Select bit3
;
;    RC0   Used by Timer1
;    RC1   Used by Timer1
;    RC2   -> DAC !CS
;    RC3   -> DAC SCLK
;    RC4   -> MUXE
;    RC5   -> DAC DIN
;    RC6
;    RC7   <- MIDI IN
;
;    RD0-7 -> NC0-7
;
;    RE0   -> !Gate 1
;    RE1   -> !Gate 2
;    RE2   -> !Gate 3
;                                                                  
;                                                                     
;*********************************************************************


	list      p=16f877A	; List directive to define processor
	#include <P16F877A.INC>
	#include <const.inc>	; Constants + Memory map
	#include <misc.inc>	; Various macros
	
	
	__CONFIG _CP_OFF & _DEBUG_OFF & _WRT_OFF & _CPD_OFF & _LVP_OFF & _BODEN_ON & _PWRTE_ON & _WDT_OFF & _HS_OSC

; '__CONFIG' directive is used to embed configuration data within .asm file.
; The lables following the directive are located in the respective .inc file.
; See respective data sheet for additional information on configuration word.




;**********************************************************************


	org     0x0000		; Processor reset vector
	clrf    PCLATH		; Ensure page bits are cleared
	goto    Init		; Go to beginning of program


	org     0x0004		; Interrupt vector location
	
	movwf   W_TEMP		; Save W, STATUS, PCLATH and FSR
	swapf	STATUS,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS_TEMP
	movfw   PCLATH
	movwf	PCLATH_TEMP
	movfw	FSR
	movwf	FSR_TEMP
	
	clrf    PCLATH		; Clear page bits
	

; Call the interupt handlers

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	INTCON,TMR0IF
	call	UpdateCV

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	PIR1,RCIF
	call	MIDIByteReceived

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	PIR1,TMR1IF
	call	Timer1Overflow

; --
	movfw	FSR_TEMP	; Restore W, STATUS, PCLATH and FSR
	movwf	FSR
	movfw	PCLATH_TEMP
	movwf	PCLATH
	swapf	STATUS_TEMP,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS
	swapf	W_TEMP,f	; Use swapf, not movf so not to disturb Z flag
	swapf	W_TEMP,w	; Swap nibbles again to get them correct
InterruptEnd
	retfie			; Return from interrupt


;**********************************************************************


Init

;
; Setup the I/O and SFR
; 

; Bank 0 Setup
	bcf	STATUS,RP0
	bcf	STATUS,RP1
	
	clrf	PORTA		; Set all outputs to off initially
	clrf	PORTB
	clrf	PORTC
	clrf	PORTD
	movlw	b'00000111'
	movwf	PORTE
	
	clrf	PIR1		; Clear all interrupt flags
	clrf	PIR2
	
	movlw	b'00000000'	; A/D clock FOSC/2, Select RA0, A/D Off
	movwf	ADCON0

	movlw	b'00111001'	; TMR1 pre-scale 1:8, TMR1 osc enabled, Internal clock, TMR1 running
	movwf	T1CON
	
	movlw	b'10000000'	; USART: Enabled, 8bit reception, Reception (CREN) disabled, Address detection disabled
	movwf	RCSTA
	
	movlw	b'00000000'	; SPI: Disable serial port, Clock idle low, SPI mode @ 5MHz
	movwf	SSPCON
	
	; Initialise general data memory
	
	clrf	MIDI_FLAGS	; Clear MIDI flags
	clrf	MIDI_MSG1	; Clear MIDI status buffer
	
	movlw	low MIDI_MSG_BUFFER	; Set pointers to MIDI ring buffer
	movwf	MIDI_MSG_PROCESS_PTR
	movwf	MIDI_MSG_INCOMING_PTR
	
	clrf	EEPROM_DATA_WRITE

	movlw	0x10		; Set pointer to just above top of note + velocity stacks
	movwf	NOTE_VELOCITY_STACK_PTR
	
	
; Bank 1 Setup
	bsf	STATUS,RP0
	bcf	STATUS,RP1
	
	movlw	b'10001110'	; Right justified, RA0 analog, RA1-7 digital, Internal Vrefs
	movwf	ADCON1
	
	movlw	b'00000001'	; Port A (RA0 input)
	movwf	TRISA
	
	movlw	b'11110000'	; Port B (RB4-7 inputs)
	movwf	TRISB
	
	movlw	b'10000000'	; Port C (RC7 input)
	movwf	TRISC
	
	movlw	b'00000000'	; Port D (all outputs)
	movwf	TRISD
	
	movwf	TRISE		; Port E (all outputs)
	
	movlw	b'00001000'	; PORTB pullups enabled, Internal TMR0 clock, Pre-scaler assigned to WDT
	movwf	OPTION_REG	; TMR0 with no pre-scale gives update rate of 51.2us per channel, 768us overall = 1.302 KHz sample rate
				; ISR with no other code (just updating CV output) takes 23us
				; S&H has 51.2-7.6 = 43.6us to settle (needs ~20us)
	
	movlw	b'01100000'	; Interupts: Global off, Peripheral on, TMR0 on, External off, RB off
	movwf	INTCON
	
	movlw	b'00100001'	; Peripheral interrupts 1: USART on, TMR1 on everything else off
	movwf	PIE1
	
	movlw	b'00000000'	; Peripheral interrupts 2: All off
	movwf	PIE2
	
	movlw	b'00000000'	; USART: Transmit disabled, Asynchronous, Low Baud Rate
	movwf	TXSTA
	
	movlw	D'9'		; USART: Baud rate 31,250
	movwf	SPBRG
	
	movlw	b'01000000'	; SPI: Transmit on rising edge
	movwf	SSPSTAT

	
	; Initialise general data memory

	movlw	1
	movwf	NEXT_CV		; Update CV1 next
	
	bcf	STATUS,IRP	; Set all CV levels to 0V (0x2D25)
	movlw	D'15'
	movwf	COUNTER
	movlw	low CV1_VALUE_L
	movwf	FSR
Init_CV_Loop
	movlw	0x25
	movwf	INDF
	incf	FSR
	movlw	0x2D
	movwf	INDF
	incf	FSR
	decfsz	COUNTER
	goto	Init_CV_Loop
	
	clrf	LFO_ANGLE_L	; Start LFO angle at 0
	clrf	LFO_ANGLE_H


; Bank 2 Setup
	bcf	STATUS,RP0
	bsf	STATUS,RP1

	clrf	vPARAMS_CHANGED		; Clear parameters changed flags
	
	; Load settings from EEPROM
	
	bsf	STATUS,RP0		; Bank 3
	bcf	EECON1,EEPGD		; Point to EEPROM data memory
	bcf	STATUS,RP0		; Bank 2
	
	movlw	EEPROM_FLAGS		; First read EEPROM_FLAGS
	movwf	EEADR
	bsf	STATUS,RP0		; Bank 3
	bsf	EECON1,RD		; Start read operation
	bcf	STATUS,RP0		; Bank 2
	movfw	EEDATA
	andlw	0x1
	movwf	TEMP1			; TEMP1 contains EEPROM_VALID in bit0
	
	btfsc	TEMP1,EEPROM_VALID	; Need to initialise EEPROM with default values?
	call	Init_InitialiseEEPROM
	
	bsf	STATUS,IRP		; Set IRP for indirect writing to bank 2
	movlw	low ParameterValuesBase	; Point to first parameter
	movwf	FSR
	
	movlw	NumberOfParameters	; Number of parameters to load
	movwf	COUNTER
	
	movlw	EEPROM_ParameterValuesBase	; Point to base parameter address in EEPROM
	movwf	EEADR
	
Init_EEPROM_Load_Loop
	; Read byte from EEPROM
	bsf	STATUS,RP0		; Bank 3
	bsf	EECON1,RD		; Start read operation
	bcf	STATUS,RP0		; Bank 2
	movfw	EEDATA
	
	movwf	INDF			; Write value to memory
	
	incf	EEADR
	incf	FSR
	decfsz	COUNTER			; Decrement and test counter
	goto	Init_EEPROM_Load_Loop
	goto	Init_EEPROM_Loaded
	

Init_InitialiseEEPROM
	bsf	STATUS,RP0		; Bank 3
	
	movlw	NumberOfParameters	; Number of parameters to initialise
	movwf	COUNTER
	
	movlw	EEPROM_ParameterValuesBase	; Point to base parameter address in EEPROM
	movwf	MTEMP1
	
	clrf	TEMP1			; Offset in default value table
	
Init_InitialiseEEPROM_Loop
	GVFT	DefaultValues,TEMP1
	movwf	MTEMP2
	call	WriteByteToEEPROM_Init
	
	incf	MTEMP1			; Incremenet EEPROM pointer and value table offset
	incf	TEMP1
	decfsz	COUNTER			; Decrement and test counter
	goto	Init_InitialiseEEPROM_Loop
	
	clrf	MTEMP1			; Set EEPROM_VALID flag
	clrf	MTEMP2
	call	WriteByteToEEPROM_Init
	
	bcf	STATUS,RP0		; Bank 2
	return
	
	
; Subroutine to get value (used by macro GVFT)
; PCLATH must be already set
GetValue
	movwf	PCL

; Default parameter values used to initialise EEPROM on first load
DefaultValues
	retlw	D'60'	; vPITCH_BEND
	retlw	D'64'	; vLFO_SPEED
	retlw	D'0'	; vLFO_EXTRA_SPEED
	retlw	D'0'	; vLFO_WAVE
	retlw	D'0'	; vLFO_DEPTH
	retlw	D'0'	; vLEGATO_RETRIG


Init_EEPROM_Loaded

	; Make Pitch Bend and LFO pre-calculations
	
	movlw	high PBR_Set		; Select program bank 2
	movwf	PCLATH
	
	; Call in RAM bank 2
	; Returns in RAM bank 0
	call	PBR_Set
	
	bsf	STATUS,RP1		; RAM bank 2 Select
	
	; Call in RAM bank 2
	; Returns in RAM bank 0
	call	LFOSpeed_Set
	
	bsf	STATUS,RP1		; RAM bank 2 Select
	
	; Call in RAM bank 2
	; Returns in RAM bank 0
	call	LFODepth_Set
	
	clrf	PCLATH			; Re-select program bank 0


; Bank 3 Setup
; 	bsf	STATUS,RP0
; 	bsf	STATUS,RP1
	
	; Nothing to do in bank 3
	
	
;--
	clrf	STATUS		; Bank 0 Select

	bsf	PORTC,2		; Set !CS high
	bsf	SSPCON,SSPEN	; SPI: Enable serial port
	bsf	RCSTA,CREN	; Enable MIDI receive
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	bsf	INTCON,GIE	; Enable interrupts


;#######################################################################

; RAM Bank 0 !
Main
	btfsc	EEPROM_DATA_WRITE,6	; Once every 6.7 seconds check if anything needs writing to EEPROM
	goto	CheckWriteToEEPROM
	
	movfw	MIDI_MSG_PROCESS_PTR	; Check if MIDI buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jz	Main
	
Main_GotMIDI
	; Got some MIDI to process ..
	movlw	high ProcessMIDI
	movwf	PCLATH			; Select program bank 2, and ready PCLATH for jump table
	call	ProcessMIDI
	clrf	PCLATH
	
	goto	Main
	
	
CheckWriteToEEPROM
	bsf	STATUS,RP1		; Bank 2 Select
	movfw	vPARAMS_CHANGED
	jz	CheckWriteToEEPROM_End	; If nothing to write, jump to end
	
	btfsc	vPARAMS_CHANGED,bPITCH_BEND	; Need to write vPITCH_BEND ?
	call	WriteToEEPROM_PITCH_BEND
	
	bcf	STATUS,RP1		; Bank 0 Select
	movfw	MIDI_MSG_PROCESS_PTR	; Check if MIDI buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jnz	Main_GotMIDI
	bsf	STATUS,RP1		; Bank 2 Select
	
	btfsc	vPARAMS_CHANGED,bLFO_SPEED	; Need to write vLFO_SPEED ?
	call	WriteToEEPROM_LFO_SPEED
	
	bcf	STATUS,RP1		; Bank 0 Select
	movfw	MIDI_MSG_PROCESS_PTR	; Check if MIDI buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jnz	Main_GotMIDI
	bsf	STATUS,RP1		; Bank 2 Select
	
	btfsc	vPARAMS_CHANGED,bLFO_EXTRA_SPEED	; Need to write vLFO_EXTRA_SPEED ?
	call	WriteToEEPROM_LFO_EXTRA_SPEED
	
	bcf	STATUS,RP1		; Bank 0 Select
	movfw	MIDI_MSG_PROCESS_PTR	; Check if MIDI buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jnz	Main_GotMIDI
	bsf	STATUS,RP1		; Bank 2 Select
	
	btfsc	vPARAMS_CHANGED,bLFO_WAVE	; Need to write vLFO_WAVE ?
	call	WriteToEEPROM_LFO_WAVE
	
	bcf	STATUS,RP1		; Bank 0 Select
	movfw	MIDI_MSG_PROCESS_PTR	; Check if MIDI buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jnz	Main_GotMIDI
	bsf	STATUS,RP1		; Bank 2 Select
	
	btfsc	vPARAMS_CHANGED,bLFO_DEPTH	; Need to write vLFO_DEPTH ?
	call	WriteToEEPROM_LFO_DEPTH
	
	bcf	STATUS,RP1		; Bank 0 Select
	movfw	MIDI_MSG_PROCESS_PTR	; Check if MIDI buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jnz	Main_GotMIDI
	bsf	STATUS,RP1		; Bank 2 Select
	
	btfsc	vPARAMS_CHANGED,bLEGATO_RETRIG	; Need to write vLEGATO_RETRIG ?
	call	WriteToEEPROM_LEGATO_RETRIG
	
CheckWriteToEEPROM_End
	bcf	STATUS,RP1		; Bank 0 Select
	clrf	EEPROM_DATA_WRITE	; Reset counter
	goto	Main
	
	
WriteParamToEEPROM	macro	vParam
	movlw	low vParam
	movwf	MTEMP1
	movfw	vParam
	movwf	MTEMP2
	bsf	STATUS,RP0		; Bank 3 Select
	call	WriteByteToEEPROM
	bcf	STATUS,RP0		; Bank 2 Select
endm
	
WriteToEEPROM_PITCH_BEND
	WriteParamToEEPROM	vPITCH_BEND
	bcf	vPARAMS_CHANGED,bPITCH_BEND
	return
	
WriteToEEPROM_LFO_SPEED
	WriteParamToEEPROM	vLFO_SPEED
	bcf	vPARAMS_CHANGED,bLFO_SPEED
	return
	
WriteToEEPROM_LFO_EXTRA_SPEED
	WriteParamToEEPROM	vLFO_EXTRA_SPEED
	bcf	vPARAMS_CHANGED,bLFO_EXTRA_SPEED
	return
	
WriteToEEPROM_LFO_WAVE
	WriteParamToEEPROM	vLFO_WAVE
	bcf	vPARAMS_CHANGED,bLFO_WAVE
	return
	
WriteToEEPROM_LFO_DEPTH
	WriteParamToEEPROM	vLFO_DEPTH
	bcf	vPARAMS_CHANGED,bLFO_DEPTH
	return
	
WriteToEEPROM_LEGATO_RETRIG
	WriteParamToEEPROM	vLEGATO_RETRIG
	bcf	vPARAMS_CHANGED,bLEGATO_RETRIG
	return
	
	
; *******************************************************************
; *******************************************************************
; *******************************************************************

	
	
; Must be call with RAM Bank 3 selected
; Returns with RAM Bank 3 selected
; MTEMP1 = Address to write to
; MTEMP2 = Value to write
WriteByteToEEPROM
	; Write byte to EEPROM
	btfsc	EECON1,WR	; Wait for previous write to finish
	goto	$-1
	
	bcf	STATUS,RP0	; RAM Bank 2
	
	movfw	MTEMP1		; Address to write to
	movwf	EEADR
	movfw	MTEMP2		; Value to write
	movwf	EEDATA
	
	bsf	STATUS,RP0	; RAM Bank 3
	
	bcf	EECON1,EEPGD	; Point to data memory
	bsf	EECON1,WREN	; Enable writes
	bcf	INTCON,GIE	; Disable interrupts
	
	movlw	0x55		; Write 0x55 to EECON2
	movwf	EECON2
	movlw	0xAA		; Write 0xAA to EECON2
	movwf	EECON2
	bsf	EECON1,WR	; Set WR bit (starts write operation)
	
	bsf	INTCON,GIE	; Enable interrupts
	bcf	EECON1,WREN	; Disable writes
	
	return
	
; Must be call with RAM Bank 3 selected
; Returns with RAM Bank 3 selected
; MTEMP1 = Address to write to
; MTEMP2 = Value to write
; This version of this subroutine assumes that interrupts are already disabled when called
WriteByteToEEPROM_Init
	; Write byte to EEPROM
	btfsc	EECON1,WR	; Wait for previous write to finish
	goto	$-1
	
	bcf	STATUS,RP0	; RAM Bank 2
	
	movfw	MTEMP1		; Address to write to
	movwf	EEADR
	movfw	MTEMP2		; Value to write
	movwf	EEDATA
	
	bsf	STATUS,RP0	; RAM Bank 3
	
	bcf	EECON1,EEPGD	; Point to data memory
	bsf	EECON1,WREN	; Enable writes
	
	movlw	0x55		; Write 0x55 to EECON2
	movwf	EECON2
	movlw	0xAA		; Write 0xAA to EECON2
	movwf	EECON2
	bsf	EECON1,WR	; Set WR bit (starts write operation)
	
	bcf	EECON1,WREN	; Disable writes
	
	return

;#######################################################################



; --------------------------------


UpdateCV
	bsf	STATUS,RP0	; Select RAM bank 1
	
	; Jump to:
	; CV1 (0001)  -> 0x0800 (00001 000  0 0000000)
	; CV2 (0010)  -> 0x0880 (00001 000  1 0000000)
	; CV3 (0011)  -> 0x0900 (00001 001  0 0000000)
	; CV4 (0100)  -> 0x0980 (00001 001  1 0000000)
	; CV5 (0101)  -> 0x0A00 (00001 010  0 0000000)
	; CV6 (0110)  -> 0x0A80 (00001 010  1 0000000)
	; CV7 (0111)  -> 0x0B00 (00001 011  0 0000000)
	; CV8 (1000)  -> 0x0B80 (00001 011  1 0000000)
	; CV9 (1001)  -> 0x0C00 (00001 100  0 0000000)
	; CV10 (1010) -> 0x0C80 (00001 100  1 0000000)
	; CV11 (1011) -> 0x0D00 (00001 101  0 0000000)
	; CV12 (1100) -> 0x0D80 (00001 101  1 0000000)
	; CV13 (1101) -> 0x0E00 (00001 110  0 0000000)
	; CV14 (1110) -> 0x0E80 (00001 110  1 0000000)
	; CV15 (1111) -> 0x0F00 (00001 111  0 0000000)
	
	decf	NEXT_CV,w		; Make CV pointer 0 based
	movwf	PCLATH			; Set into PCLATH
	rrf	PCLATH			; Rotate right
	bsf	PCLATH,3		; Set bit 3
	clrf	TEMP1
	btfsc	STATUS,C		; Carry last bit into w
	bsf	TEMP1,7
	
	movfw	TEMP1
	movwf	PCL			; Jump to routine

; --------------------------------


PointToNextIncomingBufferByte	macro
	incf	FSR			; Increment pointer
	btfss	FSR,4			; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	FSR
	
	movfw	FSR			; Check for ring buffer overrun
	subwf	MIDI_MSG_PROCESS_PTR,w
	jz	MIDIByteReceived_RingBuffOver
endm

MIDIByteReceived_RingBuffOver
	clrf	MIDI_MSG1		; Clear status byte and flags
	clrf	MIDI_FLAGS
	return


; RAM bank 0
MIDIByteReceived

	movfw	RCSTA			; Read and save and error code
	movwf	TEMP2
	
	movfw	RCREG			; Read MIDI byte
	movwf	TEMP1			; Save in TEMP1
	
	btfsc	TEMP2,OERR		; Check for overrun error
	goto	MIDIByteReceived_Error
	btfsc	TEMP2,FERR		; Check for framing error
	goto	MIDIByteReceived_Error
	
	
	btfss	TEMP1,7			; Header or data byte?
	goto	MIDIByteReceived_DataByte
	
; header byte
	movlw	b'11110000'		; System message
	andwf	TEMP1,w
	sublw	b'11110000'
	jz	MIDIByteReceived_System
	
	; Voice message
	clrf	MIDI_FLAGS		; Clear flags
	
	movfw	TEMP1			; Store status byte
	movwf	MIDI_MSG1
	
	movlw	b'11100000'		; 2 byte messages (Program Change + Aftertouch)
	andwf	TEMP1,w
	sublw	b'11000000'
	sz				; If not 2 byte message, set flag
	bsf	MIDI_FLAGS,MSG_LEN
	
	return
	
MIDIByteReceived_System
	btfss	TEMP1,3			; System common message?
	goto	MIDIByteReceived_SystemCommon
	
	; System realtime message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write byte
	movwf	INDF
	
	return
	
MIDIByteReceived_SystemCommon
	clrf	MIDI_MSG1		; Clear status buffer
	
	return				; Ignore system common messages
	

MIDIByteReceived_DataByte
	movfw	MIDI_MSG1		; Drop byte if status is 0
	rz
	
	btfsc	MIDI_FLAGS,MSG_LEN	; Test if expecting 3 byte message
	goto	MIDIByteReceived_DataByte_3byte
	
	; 2 byte message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write data byte
	movwf	INDF
	
	return
	
MIDIByteReceived_DataByte_3byte
	btfsc	MIDI_FLAGS,WHICH_BYTE	; Test if this is the 2nd data byte
	goto	MIDIByteReceived_DataByte_3byte2
	
	; 3 byte message, 1st data byte
	movfw	TEMP1			; Store data byte
	movwf	MIDI_MSG2
	
	bsf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 2nd data byte
	
	return
	
MIDIByteReceived_DataByte_3byte2
	; 3 byte message, 2nd data byte
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG2		; Write 1st data byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write 2nd data byte
	movwf	INDF
	
	bcf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 1st data byte
	
	return
	
	
MIDIByteReceived_Error
	bcf	RCSTA,CREN		; Reset receive logic / clear error flag
	bsf	RCSTA,CREN
	return
	

; --------------------------------


Timer1Overflow
	incf	EEPROM_DATA_WRITE	; Increment counter for EEPROM data write check
	
	bcf	PIR1,TMR1IF		; Clear timer 1 flag and return
	return

	
	
;#######################################################################


	; Program bank 1

	org	0x0800
; RAM bank 1
UpdateCV1
	call	UpdateCVFromRAM
	
	; Calc LFO
	
	; Determin LFO wave
	bcf	STATUS,RP0	; Select RAM bank 2
	bsf	STATUS,RP1
	movfw	vLFO_WAVE
	movwf	TEMP1
	bsf	STATUS,RP0	; Select RAM bank 1
	bcf	STATUS,RP1
	jz	CalcLFO_Sine
	decf	TEMP1
	jz	CalcLFO_Triangle
	decf	TEMP1
	jz	CalcLFO_FallingSaw
	decf	TEMP1
	jz	CalcLFO_RisingSaw
	goto	CalcLFO_Square
	
	
CalcLFO_Sine
	; Copy LFO_ANGLE_H, LFO_ANGLE_L -> TEMP2, TEMP1
	movfw	LFO_ANGLE_H
	movwf	TEMP2
	movfw	LFO_ANGLE_L
	movwf	TEMP1
	
	; if ANGLE in 2nd or 4th quadrant (bit 14 set), complement ANGLE (0-ANGLE)
	btfss	LFO_ANGLE_H,6
	goto	Quadrant_1or3
	TwosComp16 TEMP1
	
Quadrant_1or3
	; use (ANGLE >> 6) to lookup SINE in table
	
	; (ANGLE >> 6) -> TEMP1
	rrf	TEMP2
	rrf	TEMP1
	rrf	TEMP2
	rrf	TEMP1
	rrf	TEMP2
	rrf	TEMP1
	rrf	TEMP2
	rrf	TEMP1
	rrf	TEMP2
	rrf	TEMP1
	rrf	TEMP2
	rrf	TEMP1
	
	; Lookup TEMP1 in SineTable -> w
	movlw	high LookupSineTable
	movwf	PCLATH
	movfw	TEMP1
	call	LookupSineTable
	
	; if ANGLE was in 3rd or 4th quadrant (bit 15 set), complement SINE (0-SINE)
	btfsc	LFO_ANGLE_H,7
	sublw	0
	
	; SINE -> DAC Value
	; (0x5A4A + PRECALC * SINE) >> 1
	
	; PRECALC * SINE -> AARGB0, AARGB1
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	BARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movfw	LFO_PRECALC
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	AARGB0
	movlw	high FXM0808S
	movwf	PCLATH
	call	FXM0808S
	
	; 0x5A4A + AARGB0, AARGB1 -> AARGB0, AARGB1
	ADD1616lit AARGB1,0x5A4A
	
	; AARGB0, AARGB1 >> 1 -> AARGB0, AARGB1
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	
	; AARGB0, AARGB1 -> CV7_VALUE_H, CV7_VALUE_L
	movfw	AARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_H
	bcf	STATUS,RP0	; Select RAM bank 0
	movfw	AARGB1
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_L
	
	
	; Increment ANGLE by STEP
	ADD1616	LFO_ANGLE_L,LFO_STEP_L
	
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return


;-----------------------------------
	org	0x0880
; RAM bank 1
UpdateCV2
	call	UpdateCVFromRAM

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
	
;-----------------------------------

	
CalcLFO_Triangle
	; Copy LFO_ANGLE_H, LFO_ANGLE_L -> TEMP2, TEMP1
	movfw	LFO_ANGLE_H
	movwf	TEMP2
	movfw	LFO_ANGLE_L
	movwf	TEMP1
	
	; ANGLE += 0x4000 (necessary to get in phase with sine LFO)
	movlw	0x40
	addwf	TEMP2
	
	; if ANGLE > 0x7FFF (bit 15 set), complement ANGLE (0-ANGLE)
	btfss	TEMP2,7
	goto	CalcLFO_Triangle_1
	TwosComp16 TEMP1
	
CalcLFO_Triangle_1
	; ANGLE -= 0x4000
	movlw	0x40
	subwf	TEMP2
	
	; ANGLE -> DAC Value
	; 0x2D25 + PRECALC * ANGLE - 0x4000 >> 8
	
	; PRECALC * ANGLE -> AARGB0, AARGB1, AARGB2
	movfw	LFO_PRECALC
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	BARGB0
	movfw	TEMP2
	movwf	AARGB0
	movfw	TEMP1
	movwf	AARGB1
	movlw	high FXM1608S
	movwf	PCLATH
	call	FXM1608S
	
	; AARGB0, AARGB1, AARGB2 >> 8 -> AARGB0, AARGB1
	
	; 0x2D25 + AARGB0, AARGB1 -> AARGB0, AARGB1
	ADD1616lit AARGB1,0x2D25
	
	; AARGB0, AARGB1 -> CV7_VALUE_H, CV7_VALUE_L
	movfw	AARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_H
	bcf	STATUS,RP0	; Select RAM bank 0
	movfw	AARGB1
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_L
	
	
	; Increment ANGLE by STEP
	ADD1616	LFO_ANGLE_L,LFO_STEP_L
	
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
CalcLFO_Square
	; Copy LFO_ANGLE_H -> TEMP2
	movfw	LFO_ANGLE_H
	movwf	TEMP2
	
	; PRECALC -> DAC value
	; 0x2D25 +/- PRECALC * 65
	
	; PRECALC * 65 -> AARGB0, AARGB1
	movfw	LFO_PRECALC
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	AARGB0
	movlw	D'65'
	movwf	BARGB0
	movlw	high FXM0707U
	movwf	PCLATH
	call	FXM0707U
	
	movlw	high $
	movwf	PCLATH
	
	; if ANGLE bit 15 clear, DAC value low
	btfss	TEMP2,7
	goto	CalcLFO_Square_Low
	
	; 0x2D25 - AARGB0, AARGB1 -> AARGB0, AARGB1
	SUB1616litX 0x2D25,AARGB1
	goto	CalcLFO_Square_1
	
CalcLFO_Square_Low
	; 0x2D25 + AARGB0, AARGB1 -> AARGB0, AARGB1
	ADD1616lit AARGB1,0x2D25
	
CalcLFO_Square_1
	; AARGB0, AARGB1 -> CV7_VALUE_H, CV7_VALUE_L
	movfw	AARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_H
	bcf	STATUS,RP0	; Select RAM bank 0
	movfw	AARGB1
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_L
	
	
	; Increment ANGLE by STEP
	ADD1616	LFO_ANGLE_L,LFO_STEP_L
	
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0900
; RAM bank 1
UpdateCV3
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------

	
CalcLFO_FallingSaw
	; ANGLE -> DAC Value
	; 0x2D25 + PRECALC * ANGLE >> 9
	
	; PRECALC * ANGLE -> AARGB0, AARGB1, AARGB2
	movfw	LFO_PRECALC
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	BARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movfw	LFO_ANGLE_H
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	AARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movfw	LFO_ANGLE_L
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	AARGB1
	movlw	high FXM1608S
	movwf	PCLATH
	call	FXM1608S
	
	; AARGB0, AARGB1, AARGB2 >> 9 -> AARGB0, AARGB1
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	
	; 0x2D25 + AARGB0, AARGB1 -> AARGB0, AARGB1
	ADD1616lit AARGB1,0x2D25
	
	; AARGB0, AARGB1 -> CV7_VALUE_H, CV7_VALUE_L
	movfw	AARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_H
	bcf	STATUS,RP0	; Select RAM bank 0
	movfw	AARGB1
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_L
	
	
	; Increment ANGLE by STEP
	ADD1616	LFO_ANGLE_L,LFO_STEP_L
	
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
	
CalcLFO_RisingSaw
	; ANGLE -> DAC Value
	; 0x2D25 - PRECALC * ANGLE >> 9
	
	; PRECALC * ANGLE -> AARGB0, AARGB1, AARGB2
	movfw	LFO_PRECALC
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	BARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movfw	LFO_ANGLE_H
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	AARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movfw	LFO_ANGLE_L
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	AARGB1
	movlw	high FXM1608S
	movwf	PCLATH
	call	FXM1608S
	
	; AARGB0, AARGB1, AARGB2 >> 9 -> AARGB0, AARGB1
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	
	; 0x2D25 - AARGB0, AARGB1 -> AARGB0, AARGB1
	SUB1616litX 0x2D25,AARGB1
	
	; AARGB0, AARGB1 -> CV7_VALUE_H, CV7_VALUE_L
	movfw	AARGB0
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_H
	bcf	STATUS,RP0	; Select RAM bank 0
	movfw	AARGB1
	bsf	STATUS,RP0	; Select RAM bank 1
	movwf	CV7_VALUE_L
	
	
	; Increment ANGLE by STEP
	ADD1616	LFO_ANGLE_L,LFO_STEP_L
	
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0980
; RAM bank 1
UpdateCV4
	call	UpdateCVFromRAM
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0A00
; RAM bank 1
UpdateCV5
	call	UpdateCVFromRAM
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0A80
; RAM bank 1
UpdateCV6
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0B00
; RAM bank 1
UpdateCV7
	call	UpdateCVFromRAM
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0B80
; RAM bank 1
UpdateCV8
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0C00
; RAM bank 1
UpdateCV9
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0C80
; RAM bank 1
UpdateCV10
	call	UpdateCVFromRAM

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0D00
; RAM bank 1
UpdateCV11
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0D80
; RAM bank 1
UpdateCV12
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0E00
; RAM bank 1
UpdateCV13
	call	UpdateCVFromRAM
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0E80
; RAM bank 1
UpdateCV14
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0F00
; RAM bank 1
UpdateCV15
; 	call	UpdateCVFromRAM		; Not currently used, so don't update CV

	clrf	NEXT_CV		; Point to CV1
	incf	NEXT_CV
	bcf	INTCON,TMR0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------

; RAM bank 1
UpdateCVFromRAM
	bcf	STATUS,IRP	; Indirectly address RAM bank 0-1
	
	movlw	low (CV1_VALUE_L - 2)	; Base address
	movwf	FSR
	movfw	NEXT_CV		; Add 2xoffset to base address
	addwf	FSR,f
	addwf	FSR,f
	
	movwf	TEMP3		; Save offset in bank 0 accessible RAM
	
	bcf	STATUS,RP0	; Select RAM bank 0
	
	movfw	INDF		; Get LSB
	movwf	TEMP1
	incf	FSR		; Get MSB
	movfw	INDF
	movwf	TEMP2
	
	addlw	0		; Clear carry flag
	rlf	TEMP1		; Shift bits 2 places to left
	rlf	TEMP2
	rlf	TEMP1
	rlf	TEMP2,w
	
	; Set DAC to voltage
	bcf	PORTC,DACCS	; Set !CS low
	movwf	SSPBUF		; Transmit MSB
	nop2			; Wait 9 cycles before sending next byte
	nop2
	nop2
	nop
	bcf	MUXE		; While waiting: Turn off MUX (takes max 300ns)  ; Maybe move this to earlier to eliminate digital feedthrough from the SPI signal???
	movfw	TEMP1		; While waiting: Load LSB into w
	movwf	SSPBUF		; Transmit LSB and sub-bits
	nop2			; Wait 8 cycles
	rlf	TEMP3		; While waiting: Shift MUX bits to left
	btfsc	TEMP3,4		; While waiting: Copy bit 4 to left
	bsf	TEMP3,5
	movfw	TEMP3
	movwf	PORTA		; While waiting: Select MUX channel
	nop
	bsf	PORTC,DACCS	; Set !CS high (loads DAC)
	
	wait_us 5		; Wait 5us for DAC and opamps to settle
				; Maybe less wait with faster opamp?
	
	bsf	MUXE		; Turn on MUX (takes typ 105ns, max 600ns)
	
	; S&H now has until next call of UpdateCVFromRAM to settle

	bsf	STATUS,RP0	; Select RAM bank 1
	
	return


;#######################################################################


	; Program bank 2
	
	org	0x1000
	
	; Implemented messages:
	;
	; 1000cccc nnnnnnnn vvvvvvvv  Note Off
	; 1001cccc nnnnnnnn vvvvvvvv  Note On
	; 1011cccc cccccccc vvvvvvvv  Control Change
	; 1110cccc llllllll hhhhhhhh  Pitch Bend

; RAM Bank 0
PointToNextProcessBufferByte
	incf	MIDI_MSG_PROCESS_PTR	; Increment pointer
	btfss	MIDI_MSG_PROCESS_PTR,4	; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	MIDI_MSG_PROCESS_PTR
	return
	

;***********************************************************************
;***********************************************************************

; Parse MIDI message here
; RAM Bank 0
ProcessMIDI
	call	PointToNextProcessBufferByte
	
	bcf	STATUS,IRP	; Indirectly address banks 0-1
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	movfw	INDF		; Get and save MIDI channel
	andlw	b'00001111'
	movwf	MTEMP1
	
	rrf	INDF,w		; Get message
	movwf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2,w
	andlw	b'00000111'
	
	addwf	PCL		; Jump to handler (PCLATH already set before call to ProcessMIDI)
	goto	ProcessMIDI_NoteOff
	goto	ProcessMIDI_NoteOn
	goto	ProcessMIDI_PolyAftertouch
	goto	ProcessMIDI_ControlChange
	goto	ProcessMIDI_ProgramChange
	goto	ProcessMIDI_Aftertouch
	goto	ProcessMIDI_PitchBend
	goto	ProcessMIDI_System


;***********************************************************************
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_NoteOff
	call	PointToNextProcessBufferByte	; Point FSR to note value
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	call	PointToNextProcessBufferByte	; Point MIDI_MSG_PROCESS_PTR to velocity
	
	; Check MIDI channel
	rrf	PORTB,w
	movwf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	comf	MTEMP3,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	rnz
	
	movfw	INDF		; Get and save note value
	movwf	MTEMP1
	
ProcessMIDI_NoteOff_1
	
	; Check if there are no other unreleased notes on note stack
	movfw	NOTE_VELOCITY_STACK_PTR
	addlw	0xF1		; Test if >= 0x0F
	jc	ProcessMIDI_NoteOff_LastNote
	
	; Search for this note on stack
	
	movfw	NOTE_VELOCITY_STACK_PTR
	movwf	MTEMP2		; Counter (counts up from NOTE_VELOCITY_STACK_PTR to 0x0F)
	
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	
	addlw	low NOTE_STACK	; Point to top of note stack
	movwf	FSR
	
ProcessMIDI_NoteOff_Loop
	movfw	INDF
	subwf	MTEMP1,w
	jz	ProcessMIDI_NoteOff_FoundNote
	
	incf	FSR
	incf	MTEMP2		; Loop until at bottom of stack
	btfss	MTEMP2,4
	goto	ProcessMIDI_NoteOff_Loop
	
	; Note not found in stack, so no need to remove and shift stack
	
	goto	ProcessMIDI_NoteOff_ChangeCVs	; Change CVs to values on top of stacks and return
	
	
ProcessMIDI_NoteOff_FoundNote
	movfw	MTEMP2		; If MTEMP2 == NOTE_VELOCITY_STACK_PTR, note released is at top of stack so no need to shift stack
	subwf	NOTE_VELOCITY_STACK_PTR,w
	jnz	ProcessMIDI_NoteOff_FoundNote_1
	incf	NOTE_VELOCITY_STACK_PTR		; Adjust stack pointer
	goto	ProcessMIDI_NoteOff_ChangeCVs	; Change CVs to values on top of stacks and return

ProcessMIDI_NoteOff_FoundNote_1
	; Remove note and velocity value and shift stack
	decf	FSR
	movfw	INDF
	incf	FSR
	movwf	INDF
	
	movlw	(VELOCITY_STACK - NOTE_STACK)
	addwf	FSR
	
	decf	FSR
	movfw	INDF
	incf	FSR
	movwf	INDF
	
	movlw	(1 + VELOCITY_STACK - NOTE_STACK)
	subwf	FSR
	
	decf	MTEMP2		; Loop until at top of stack
	movfw	MTEMP2
	subwf	NOTE_VELOCITY_STACK_PTR,w
	jnz	ProcessMIDI_NoteOff_FoundNote_1

ProcessMIDI_NoteOff_FoundNote_2
	incf	NOTE_VELOCITY_STACK_PTR		; Adjust stack pointer
	goto	ProcessMIDI_NoteOff_ChangeCVs	; Change CVs to values on top of stacks and return
	
	
ProcessMIDI_NoteOff_ChangeCVs
	; Read next note and velocity value off stack tops and update CVs
	movfw	NOTE_VELOCITY_STACK_PTR
	addlw	low NOTE_STACK
	movwf	FSR
	movfw	INDF
	call	NoteToCV		; Note (in w) -> CV
	movfw	NOTE_VELOCITY_STACK_PTR
	addlw	low VELOCITY_STACK
	movwf	FSR
	movfw	INDF
	call	VelocityToCV		; Velocity (in w) -> CV
	
	return
	
	
ProcessMIDI_NoteOff_LastNote
	movlw	0x10		; Reset note stack pointer
	movwf	NOTE_VELOCITY_STACK_PTR
	
	bsf	PORTE,GATE1	; Gate low
	
	; Set spare CV5 low (for external gates)
	bsf	STATUS,RP0	; Bank 1 Select
	movlw	0x2D
	bcf	INTCON,GIE	; Disable interrupts while changing CV RAM
	movwf	CV5_VALUE_H
	movlw	0x25
	movwf	CV5_VALUE_L
	bsf	INTCON,GIE	; Enable interrupts again
	bcf	STATUS,RP0	; Bank 0 Select
	
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_NoteOn
	call	PointToNextProcessBufferByte	; Point FSR to note value
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	call	PointToNextProcessBufferByte	; Point MIDI_MSG_PROCESS_PTR to velocity
	
	; Check MIDI channel
	rrf	PORTB,w
	movwf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	comf	MTEMP3,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	rnz
	
	movfw	INDF			; Get and save note value
	movwf	MTEMP1
	
	movfw	MIDI_MSG_PROCESS_PTR	; Point FSR to velocity value
	movwf	FSR
	movfw	INDF			; Get and save velocity value
	movwf	MTEMP2
	jz	ProcessMIDI_NoteOff_1	; If velocity=0, count as note off
	
	; Velocity (in w) -> CV
	call	VelocityToCV
	
	; Note -> CV
	movfw	MTEMP1
	call	NoteToCV
	
	
	; Save note and velocity on stacks
	
	movf	NOTE_VELOCITY_STACK_PTR,f	; Test if stack pointer=0
	jz	ProcessMIDI_NoteOn_1		; If so, no stack left so don't push note
	
	decf	NOTE_VELOCITY_STACK_PTR	; Decrement stack pointer
	
	bsf	STATUS,IRP		; Indirectly address banks 2-3

	movfw	NOTE_VELOCITY_STACK_PTR
	addlw	low NOTE_STACK		; Point to note stack byte
	movwf	FSR
	
	movfw	MTEMP1			; Push note value onto stack
	movwf	INDF
	
	movfw	NOTE_VELOCITY_STACK_PTR
	addlw	low VELOCITY_STACK	; Point to velocity stack byte
	movwf	FSR
	
	movfw	MTEMP2			; Push velocity value onto stack
	movwf	INDF
	
ProcessMIDI_NoteOn_1
	
	btfsc	PORTE,GATE1	; If gate already off, no need to re-trigger
	goto	ProcessMIDI_NoteOn_GateOn
	
	bsf	STATUS,RP1	; Bank 2 Select
	movf	vLEGATO_RETRIG,f	; If legato retrigger mode is off, no need to re-trigger
	bcf	STATUS,RP1	; Bank 0 Select
	jz	ProcessMIDI_NoteOn_GateOn
	
	; Turn off gates momentarily before turning on again
	bsf	PORTE,GATE1	; Gate low
	
	; Set spare CV5 low (for external gates)
	bsf	STATUS,RP0	; Bank 1 Select
	movlw	0x2D
	bcf	INTCON,GIE	; Disable interrupts while changing CV RAM
	movwf	CV5_VALUE_H
	movlw	0x25
	movwf	CV5_VALUE_L
	bsf	INTCON,GIE	; Enable interrupts again
	bcf	STATUS,RP0	; Bank 0 Select
	
	wait_ms_M 1		; Wait 1ms with gate low

ProcessMIDI_NoteOn_GateOn
	bcf	PORTE,GATE1	; Gate high
	
	; Set spare CV5 to 5V (for external gates)
	bsf	STATUS,RP0	; Bank 1 Select
	movlw	0x17
	bcf	INTCON,GIE	; Disable interrupts while changing CV RAM
	movwf	CV5_VALUE_H
	movlw	0xD1
	movwf	CV5_VALUE_L
	bsf	INTCON,GIE	; Enable interrupts again
	bcf	STATUS,RP0	; Bank 0 Select
	
	return
	
	
	
; Note value in w
; RAM bank 0
NoteToCV
	bsf	STATUS,RP0	; Bank 1 Select
	
	; Note -> DAC value
	; (127 - N) * 0x5B
	
	; 127 - N -> w
	sublw	D'127'
	
	; w * 0x5B -> AARGB0,AARGB1
	movwf	AARGB0
	movlw	0x5B
	movwf	BARGB0
	call	FXM0707U
	
	; Write to CV1
	movfw	AARGB0
	bcf	INTCON,GIE	; Disable interrupts while changing CV RAM
	movwf	CV1_VALUE_H
	movwf	CV2_VALUE_H	; Also update spare CV2 so it can be used for external VCOs
	movfw	AARGB1
	movwf	CV1_VALUE_L
	movwf	CV2_VALUE_L	; Also update spare CV2 so it can be used for external VCOs
	bsf	INTCON,GIE	; Enable interrupts again
	bcf	STATUS,RP0	; Bank 0 Select
	
	return
	
; Velocity value in w
; RAM bank 0
VelocityToCV
	bsf	STATUS,RP0	; Bank 1 Select
	
	; Velocity -> DAC value
	; (127 - N) * 0x5B
	
	; 127 - N -> w
	sublw	D'127'
	
	; w * 0x5B -> AARGB0,AARGB1
	movwf	AARGB0
	movlw	0x5B
	movwf	BARGB0
	call	FXM0707U
	
	; Write to CV10
	movfw	AARGB0
	bcf	INTCON,GIE	; Disable interrupts while changing CV RAM
	movwf	CV10_VALUE_H
	movfw	AARGB1
	movwf	CV10_VALUE_L
	bsf	INTCON,GIE	; Enable interrupts again
	bcf	STATUS,RP0	; Bank 0 Select	
	
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_PolyAftertouch
	call	PointToNextProcessBufferByte
	call	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_ControlChange
	call	PointToNextProcessBufferByte	; Get and save controller number
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	movwf	MTEMP2
	call	PointToNextProcessBufferByte	; Get and save controller value
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	movwf	MTEMP4
	
	; Check MIDI channel
	rrf	PORTB,w
	movwf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	comf	MTEMP3,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	rnz
	
	movfw	MTEMP2
	sublw	D'1'		; Mod wheel -> MOD
	jz	ProcessMIDI_ControlChange_MOD
	
	movfw	MTEMP2
	sublw	D'65'		; Portamento On/Off -> SLIDE
	jz	ProcessMIDI_ControlChange_SLIDE
	
	movfw	MTEMP2
	sublw	D'80'		; General Purpose Button 1 -> ACCENT
	jz	ProcessMIDI_ControlChange_ACCENT
	
	movfw	MTEMP2
	sublw	D'120'		; All sound Off
	jz	ProcessMIDI_NoteOff_LastNote
	
	movfw	MTEMP2
	sublw	D'123'		; All notes off
	jz	ProcessMIDI_NoteOff_LastNote
	
	movfw	MTEMP2
	sublw	D'16'		; General Purpose Slider 1 -> Pitch Bend Range
	jz	ProcessMIDI_ControlChange_PBR
	
	movfw	MTEMP2
	sublw	D'17'		; General Purpose Slider 2 -> LFO Speed
	jz	ProcessMIDI_ControlChange_LFOSpeed
	
	movfw	MTEMP2
	sublw	D'18'		; General Purpose Slider 3 -> LFO Extra Speed
	jz	ProcessMIDI_ControlChange_LFOExtraSpeed
	
	movfw	MTEMP2
	sublw	D'19'		; General Purpose Slider 4 -> LFO Waveform
	jz	ProcessMIDI_ControlChange_LFOWave
	
	movfw	MTEMP2
	sublw	D'92'		; Tremolo Level -> LFO Depth
	jz	ProcessMIDI_ControlChange_LFODepth
	
	movfw	MTEMP2
	sublw	D'82'		; General Purpose Button 3 -> Legato Retrigger
	jz	ProcessMIDI_ControlChange_LegRetrig
	
	return
	
	
ProcessMIDI_ControlChange_MOD
	bsf	STATUS,RP0	; Bank 1 Select
	
	movfw	MTEMP4

	; Mod Wheel -> DAC value
	; (127 - N) * 0x5B
	
	; 127 - N -> w
	sublw	D'127'
	
	; w * 0x5B -> AARGB0,AARGB1
	movwf	AARGB0
	movlw	0x5B
	movwf	BARGB0
	call	FXM0707U
	
	; Write to CV10
	movfw	AARGB0
	bcf	INTCON,GIE	; Disable interrupts while changing CV RAM
	movwf	CV13_VALUE_H
	movfw	AARGB1
	movwf	CV13_VALUE_L
	bsf	INTCON,GIE	; Enable interrupts again
	bcf	STATUS,RP0	; Bank 0 Select	
	
	return


ProcessMIDI_ControlChange_SLIDE
	btfsc	MTEMP4,6	; CC value >= 64 ?
	goto	SLIDE_Hi
	
	bsf	PORTE,GATE2	; Gate low
	return
	
SLIDE_Hi
	bcf	PORTE,GATE2	; Gate high
	return


ProcessMIDI_ControlChange_ACCENT
	btfsc	MTEMP4,6	; CC value >= 64 ?
	goto	ACCENT_Hi
	
	bsf	PORTE,GATE3	; Gate low
	return
	
ACCENT_Hi
	bcf	PORTE,GATE3	; Gate high
	return
	

ProcessMIDI_ControlChange_PBR
	bsf	STATUS,RP1	; Bank 2 Select
	
	movfw	MTEMP4
	movwf	vPITCH_BEND
	
	bsf	vPARAMS_CHANGED,bPITCH_BEND	; Flag that value needs to be written to EEPROM
	
	addlw	D'255' - D'120'	; CC value <= 120 ?
	jnc	PBR_Set
	
	movlw	D'120'		; If not, set to 120
	movwf	vPITCH_BEND
	
	bsf	vPARAMS_CHANGED,bPITCH_BEND	; Flag that value needs to be written to EEPROM

; Call in RAM bank 2
; Returns in RAM bank 0
PBR_Set
	movfw	vPITCH_BEND
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1

	; PRECALC2 = PBR * 0x03C1 / 3
	
	; PBR * 0x03C1 -> AARGB0, AARGB1, AARGB2
	movwf	BARGB0
	movlw	0x03
	movwf	AARGB0
	movlw	0xC1
	movwf	AARGB1
	call	FXM1507U
	
	; AARGB0, AARGB1, AARGB2 / 3 -> AARGB0, AARGB1, AARGB2
	clrf	BARGB0
	movlw	3
	movwf	BARGB1
	call	FXD2315U
	
	; AARGB1, AARGB2 -> BEND_PRECALC2_H, BEND_PRECALC2_L
	movfw	AARGB1
	movwf	BEND_PRECALC2_H
	movfw	AARGB2
	movwf	BEND_PRECALC2_L

	; PRECALC1 = 0x2D25 + PBR * 0x03C1 / 3 >> 3
	
	; AARGB1, AARGB2 >> 3 -> BEND_PRECALC1_H, BEND_PRECALC1_L
	bcf	STATUS,C
	rrf	AARGB1
	rrf	AARGB2
	bcf	STATUS,C
	rrf	AARGB1
	rrf	AARGB2
	bcf	STATUS,C
	rrf	AARGB1,w
	movwf	BEND_PRECALC1_H
	rrf	AARGB2,w
	movwf	BEND_PRECALC1_L
	
	; 0x2D25 -> MTEMP2, MTEMP1
	movlw	0x2D
	movwf	MTEMP2
	movlw	0x25
	movwf	MTEMP1
	
	; 0x2D25 + BEND_PRECALC1_H, BEND_PRECALC1_L -> BEND_PRECALC1_H, BEND_PRECALC1_L
	ADD1616	BEND_PRECALC1_L,MTEMP1
	
	bcf	STATUS,RP0	; Bank 0 Select
	clrf	EEPROM_DATA_WRITE	; Reset counter for writing param change to EEPROM (to minimise EEPROM writes)
	return


ProcessMIDI_ControlChange_LFOSpeed
	bsf	STATUS,RP1	; Bank 2 Select
	
	movfw	MTEMP4
	movwf	vLFO_SPEED
	
	bsf	vPARAMS_CHANGED,bLFO_SPEED	; Flag that value needs to be written to EEPROM
	
; Call in RAM bank 2
; Returns in RAM bank 0
LFOSpeed_Set
	; Copy vLFO_EXTRA_SPEED to MTEMP1
	movfw	vLFO_EXTRA_SPEED
	movwf	MTEMP1
	; Copy vLFO_SPEED to MTEMP2
	movfw	vLFO_SPEED
	movwf	MTEMP2
	

	; LFO Speed -> STEP
	; 3276 / (128 - LFO_Speed)
	
	; (128 - LFO_Speed) -> w
	sublw	D'128'
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	; 3276 / w -> AARGB0, AARGB1
	movwf	BARGB0
	movlw	high D'3276'
	movwf	AARGB0
	movlw	low D'3276'
	movwf	AARGB1
	call	FXD1608U
	
	; AARGB0, AARGB1 -> LFO_STEP_H, LFO_STEP_L
	movfw	AARGB0
	movwf	LFO_STEP_H
	movfw	AARGB1
	movwf	LFO_STEP_L
	
	
	; if LFO_Speed == 127:
	; STEP += LFO_Extra_Speed * 16
	
	movlw	D'127'
	subwf	MTEMP2,w
	jnz	LFOSpeed_Set_End
	
	; LFO_Extra_Speed * 32 -> MTEMP2, MTEMP1
	clrf	MTEMP2
	bcf	STATUS,C
	rlf	MTEMP1
	rlf	MTEMP2
	rlf	MTEMP1
	rlf	MTEMP2
	rlf	MTEMP1
	rlf	MTEMP2
	rlf	MTEMP1
	rlf	MTEMP2
	rlf	MTEMP1
	rlf	MTEMP2
	
	; LFO_STEP_H, LFO_STEP_L + MTEMP2, MTEMP1 -> LFO_STEP_H, LFO_STEP_L
	ADD1616	LFO_STEP_L,MTEMP1
	
LFOSpeed_Set_End
	bcf	STATUS,RP0	; Bank 0 Select
	clrf	EEPROM_DATA_WRITE	; Reset counter for writing param change to EEPROM (to minimise EEPROM writes)
	return


ProcessMIDI_ControlChange_LFOExtraSpeed
	bsf	STATUS,RP1	; Bank 2 Select
	
	movfw	MTEMP4
	movwf	vLFO_EXTRA_SPEED
	
	bsf	vPARAMS_CHANGED,bLFO_EXTRA_SPEED	; Flag that value needs to be written to EEPROM
	
	goto	LFOSpeed_Set


ProcessMIDI_ControlChange_LFOWave
	movfw	MTEMP4
	addlw	D'255' - D'4'	; if CC value > 4 then ignore
	rc

	bsf	STATUS,RP1	; Bank 2 Select
	
	movfw	MTEMP4
	movwf	vLFO_WAVE
	
	bsf	vPARAMS_CHANGED,bLFO_WAVE	; Flag that value needs to be written to EEPROM.
	
	bcf	STATUS,RP1	; Bank 0 Select
	clrf	EEPROM_DATA_WRITE	; Reset counter for writing param change to EEPROM (to minimise EEPROM writes)
	return


ProcessMIDI_ControlChange_LFODepth
	bsf	STATUS,RP1	; Bank 2 Select
	
	movfw	MTEMP4
	movwf	vLFO_DEPTH
	
	bsf	vPARAMS_CHANGED,bLFO_DEPTH	; Flag that value needs to be written to EEPROM
	
; Call in RAM bank 2
; Returns in RAM bank 0
LFODepth_Set
	; PRECALC = 0x12C5 * LFO_Depth >> 12
	
	movfw	vLFO_DEPTH
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	; 0x12C5 * LFO_Depth -> AARGB0, AARGB1, AARGB2
	movwf	BARGB0
	movlw	0x12
	movwf	AARGB0
	movlw	0xC5
	movwf	AARGB1
	call	FXM1507U
	
	; AARGB0, AARGB1, AARGB2 >> 13 -> AARGB1
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB0
	rrf	AARGB1
	
	; AARGB1 -> LFO_PRECALC
	movfw	AARGB1
	movwf	LFO_PRECALC
	
	bcf	STATUS,RP0	; Bank 0 Select
	clrf	EEPROM_DATA_WRITE	; Reset counter for writing param change to EEPROM (to minimise EEPROM writes)
	return


ProcessMIDI_ControlChange_LegRetrig
	bsf	STATUS,RP1	; Bank 2 Select
	
	btfsc	MTEMP4,6	; CC value >= 64 ?
	goto	LegRetrig_On
	
	clrf	vLEGATO_RETRIG	; Off
	
	bsf	vPARAMS_CHANGED,bLEGATO_RETRIG	; Flag that value needs to be written to EEPROM.
	
	bcf	STATUS,RP1	; Bank 0 Select
	clrf	EEPROM_DATA_WRITE	; Reset counter for writing param change to EEPROM (to minimise EEPROM writes)
	return
	
LegRetrig_On
	movlw	1		; On
	movwf	vLEGATO_RETRIG
	
	bsf	vPARAMS_CHANGED,bLEGATO_RETRIG	; Flag that value needs to be written to EEPROM
	
	bcf	STATUS,RP1	; Bank 0 Select
	clrf	EEPROM_DATA_WRITE	; Reset counter for writing param change to EEPROM (to minimise EEPROM writes)
	return

	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_ProgramChange
	call	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_Aftertouch
	call	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_PitchBend
	
	; Check MIDI channel
	rrf	PORTB,w
	movwf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	comf	MTEMP3,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	jnz	ProcessMIDI_PitchBend_End
	
	; Get 8 bit pitch bend value into MTEMP1
	
	call	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	rlf	INDF		; Get LSB and rotate left 2 bits to push MSbit into C
	rlf	INDF,w
	
	call	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	rlf	INDF,w		; Get MSB and rotate left 1 bit, picking up LSbit from LSB
	
	movwf	MTEMP1		; Save 8-bit pitch bend value in MTEMP1
	
	bsf	STATUS,RP0	; Bank 1 Select
	
	movlw	D'128'		; Check if pitch bend centered
	subwf	MTEMP1,w
	jz	ProcessMIDI_PitchBend_Centered
	
	
	; DAC value = PRECALC1  -  (PB >> 6) * PRECALC2 >> 10

	; (PB >> 6) * PRECALC2  -> AARGB0, AARGB1, AARGB2
	movfw	MTEMP1		; (PB >> 6)
	movwf	BARGB0
	movfw	BEND_PRECALC2_H
	movwf	AARGB0
	movfw	BEND_PRECALC2_L
	movwf	AARGB1
	call	FXM1608U
	
	; if (PB >> 6) == 0xFF then
	; AARGB0, AARGB1 += PRECALC2_H
	; (to give perfect tuning for pitch bend fully up)
	incf	MTEMP1,w
	jnz	ProcessMIDI_PitchBend_1
	ADD1608	AARGB1,BEND_PRECALC2_H
	
ProcessMIDI_PitchBend_1
	; AARGB0, AARGB1, AARGB2 >> 10  -> AARGB0, AARGB1
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	
	; PRECALC1 -> BARGB0, BARGB1
	movfw	BEND_PRECALC1_H
	movwf	BARGB0
	movfw	BEND_PRECALC1_L
	movwf	BARGB1
	
	; BARGB0, BARGB1 - AARGB0, AARGB1 -> BARGB0, BARGB1
	SUB1616	BARGB1,AARGB1
	
	; BARGB0, BARGB1 -> CV4_VALUE_H, CV4_VALUE_L
	movfw	BARGB0
	movwf	CV4_VALUE_H
	movfw	BARGB1
	movwf	CV4_VALUE_L
	
	bcf	STATUS,RP0	; Bank 0 Select
	return
	
	
ProcessMIDI_PitchBend_Centered
	; Set CV4 to 0V (0x2D25)
	movlw	0x2D
	movwf	CV4_VALUE_H
	movlw	0x25
	movwf	CV4_VALUE_L
	
	bcf	STATUS,RP0	; Bank 0 Select
	return
	
	
ProcessMIDI_PitchBend_End
	call	PointToNextProcessBufferByte
	call	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_System
	return
	
	
;***********************************************************************
;***********************************************************************
	
	

; Maths Functions
#include <maths.inc>


;#######################################################################


	; Program bank 3
	
	org	0x1800
	
	nop
	
	org	0x18FE
	
; Sine lookup table
; There are 256 7 bit values here for the 1st quadrant, scaled from 0 to 127
LookupSineTable
	incf	PCLATH
	movwf	PCL
SineTable	
	dt	0x00,0x01,0x02,0x02,0x03,0x04,0x05,0x05
	dt	0x06,0x07,0x08,0x09,0x09,0x0A,0x0B,0x0C
	dt	0x0C,0x0D,0x0E,0x0F,0x10,0x10,0x11,0x12
	dt	0x13,0x13,0x14,0x15,0x16,0x16,0x17,0x18
	dt	0x19,0x1A,0x1A,0x1B,0x1C,0x1D,0x1D,0x1E
	dt	0x1F,0x20,0x20,0x21,0x22,0x23,0x23,0x24
	dt	0x25,0x26,0x26,0x27,0x28,0x29,0x29,0x2A
	dt	0x2B,0x2C,0x2C,0x2D,0x2E,0x2E,0x2F,0x30
	dt	0x31,0x31,0x32,0x33,0x33,0x34,0x35,0x36
	dt	0x36,0x37,0x38,0x38,0x39,0x3A,0x3A,0x3B
	dt	0x3C,0x3D,0x3D,0x3E,0x3F,0x3F,0x40,0x41
	dt	0x41,0x42,0x43,0x43,0x44,0x45,0x45,0x46
	dt	0x47,0x47,0x48,0x48,0x49,0x4A,0x4A,0x4B
	dt	0x4C,0x4C,0x4D,0x4E,0x4E,0x4F,0x4F,0x50
	dt	0x51,0x51,0x52,0x52,0x53,0x54,0x54,0x55
	dt	0x55,0x56,0x56,0x57,0x58,0x58,0x59,0x59
	dt	0x5A,0x5A,0x5B,0x5B,0x5C,0x5D,0x5D,0x5E
	dt	0x5E,0x5F,0x5F,0x60,0x60,0x61,0x61,0x62
	dt	0x62,0x63,0x63,0x64,0x64,0x65,0x65,0x66
	dt	0x66,0x66,0x67,0x67,0x68,0x68,0x69,0x69
	dt	0x6A,0x6A,0x6A,0x6B,0x6B,0x6C,0x6C,0x6D
	dt	0x6D,0x6D,0x6E,0x6E,0x6F,0x6F,0x6F,0x70
	dt	0x70,0x70,0x71,0x71,0x71,0x72,0x72,0x72
	dt	0x73,0x73,0x73,0x74,0x74,0x74,0x75,0x75
	dt	0x75,0x76,0x76,0x76,0x76,0x77,0x77,0x77
	dt	0x78,0x78,0x78,0x78,0x79,0x79,0x79,0x79
	dt	0x7A,0x7A,0x7A,0x7A,0x7A,0x7B,0x7B,0x7B
	dt	0x7B,0x7B,0x7C,0x7C,0x7C,0x7C,0x7C,0x7C
	dt	0x7D,0x7D,0x7D,0x7D,0x7D,0x7D,0x7D,0x7E
	dt	0x7E,0x7E,0x7E,0x7E,0x7E,0x7E,0x7E,0x7E
	dt	0x7E,0x7E,0x7F,0x7F,0x7F,0x7F,0x7F,0x7F
	dt	0x7F,0x7F,0x7F,0x7F,0x7F,0x7F,0x7F,0x7F

;#######################################################################


	; EEPROM
	
 	org	0x2100

; 	de	0xFF		; Set EEPROM_FLAGS,EEPROM_VALID to 'invalid'


;#######################################################################


	END			 ; directive 'end of program'


